use ezcad::pen::Pen;
use log::debug;
use serde::{Deserialize, Serialize};

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct Patch {
    color: Option<(u8, u8, u8)>,
    enabled: Option<bool>,
    loop_count: Option<u32>,
    speed: Option<f64>,
    power: Option<f64>,
    frequency: Option<u32>,
}

impl Patch {
    fn patch(&self, id: usize, target: &mut Vec<Pen>) {
        let to_patch: &mut Pen = target.get_mut(id).expect("Invalid pen index");

        if let Some(color) = self.color {
            debug!("Patching color for pen #{}: {:?}", id, color);
            *to_patch.color = color.into()
        }

        if let Some(enabled) = self.enabled {
            debug!("Patching enablement for pen #{}: {}", id, enabled);
            *to_patch.disabled = !enabled as u32;
        }

        if let Some(loop_count) = self.loop_count {
            debug!("Patching loop count for pen #{}: {}", id, loop_count);
            *to_patch.loop_count = loop_count;
        }

        if let Some(speed) = self.speed {
            debug!("Patching speed for pen #{}: {}", id, speed);
            *to_patch.speed = speed;
        }

        if let Some(power) = self.power {
            debug!("Patching power for pen #{}: {}", id, power);
            *to_patch.power = power;
        }

        if let Some(frequency) = self.frequency {
            debug!("Patching frequency for pen #{}: {}", id, frequency);
            *to_patch.frequency = frequency;
            *to_patch.frequency_2 = frequency.try_into().unwrap();
        }

        // Always enable custom settings for pen
        *to_patch.use_default = 0;
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct PatchPen {
    pen: usize,
    #[serde(flatten)]
    patch: Patch,
}

impl PatchPen {
    fn patch(&self, target: &mut Vec<Pen>) {
        self.patch.patch(self.pen, target);
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct ClonePen {
    from: usize,
    to: usize,
    #[serde(flatten)]
    patch: Option<Patch>,
}

impl ClonePen {
    fn clone(&self, target: &mut Vec<Pen>) {
        debug!("Cloning pen #{} to #{}", self.from, self.to);
        assert!(
            self.to > self.from,
            "Target pen must be greater than source pen"
        );

        // Clone pen
        let src: Pen = target.get(self.from).expect("Invalid pen index").clone();
        let dst: &mut Pen = target.get_mut(self.to).expect("Invalid pen index");
        *dst = src;

        // Patch pen if needed
        if let Some(patch) = &self.patch {
            patch.patch(self.to, target);
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub enum PatternField {
    Loops(i32),
    Speed(f64),
    Power(f64),
    Frequency(i32),
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct PatternPen {
    from: usize,
    to: usize,
    field: PatternField,
}

impl PatternPen {
    fn pattern(&self, target: &mut Vec<Pen>) {
        debug!("Patterning from pen #{} to #{}", self.from, self.to);
        assert!(
            self.to > self.from,
            "Target pen(s) must be greater than source pen"
        );

        // Obtain settings from first pen
        let src: Pen = target.get(self.from).expect("Invalid pen index").clone();

        let mut setting: PatternField = match self.field {
            PatternField::Loops(_) => {
                debug!(
                    "Initial loop count from pen #{}: {}",
                    self.from, *src.loop_count
                );
                PatternField::Loops((*src.loop_count).try_into().unwrap())
            }
            PatternField::Speed(_) => {
                debug!("Initial speed from pen #{}: {}", self.from, *src.speed);
                PatternField::Speed(*src.speed)
            }
            PatternField::Power(_) => {
                debug!("Initial power from pen #{}: {}", self.from, *src.power);
                PatternField::Power(*src.power)
            }
            PatternField::Frequency(_) => {
                debug!(
                    "Initial frequency from pen #{}: {}",
                    self.from, *src.frequency
                );
                PatternField::Frequency((*src.frequency).try_into().unwrap())
            }
        };

        for dst in (self.from..=self.to).skip(1) {
            // Clone to target pen
            let pen: &mut Pen = target.get_mut(dst).expect("Invalid pen index");
            *pen = src.clone();

            // Calculate new setting
            setting = match (setting, &self.field) {
                (PatternField::Loops(prev), PatternField::Loops(incr)) => {
                    debug!("Patching loop count for pen #{}: {}", dst, prev + incr);
                    PatternField::Loops(prev + incr)
                }
                (PatternField::Speed(prev), PatternField::Speed(incr)) => {
                    debug!("Patching speed for pen #{}: {}", dst, prev + incr);
                    PatternField::Speed(prev + incr)
                }
                (PatternField::Power(prev), PatternField::Power(incr)) => {
                    debug!("Patching power for pen #{}: {}", dst, prev + incr);
                    PatternField::Power(prev + incr)
                }
                (PatternField::Frequency(prev), PatternField::Frequency(incr)) => {
                    debug!("Patching frequency for pen #{}: {}", dst, prev + incr);
                    PatternField::Frequency(prev + incr)
                }
                _ => unreachable!(),
            };

            // Patch updated value
            match setting {
                PatternField::Loops(x) => *pen.loop_count = x.try_into().unwrap(),
                PatternField::Speed(x) => *pen.speed = x,
                PatternField::Power(x) => *pen.power = x,
                PatternField::Frequency(x) => {
                    *pen.frequency = x.try_into().unwrap();
                    *pen.frequency_2 = x.try_into().unwrap();
                }
            }

            // Always enable custom settings for pen
            *pen.use_default = 0;
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub enum PenOperation {
    Patch(PatchPen),
    Clone(ClonePen),
    Pattern(PatternPen),
}

pub trait PatchPens {
    fn patch(&self, target: &mut Vec<Pen>);
}

impl PatchPens for Vec<PenOperation> {
    fn patch(&self, target: &mut Vec<Pen>) {
        for op in self {
            match op {
                PenOperation::Patch(x) => x.patch(target),
                PenOperation::Clone(x) => x.clone(target),
                PenOperation::Pattern(x) => x.pattern(target),
            }
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct Config {
    pub pens: Vec<PenOperation>,
}
